/**
 * <h1>Commander</h1>
 *
 * <p>Process interactive debugger commands.</p>
 *
 * <p>Copyright (c) 2020 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
package backend.debugger;

import java.util.HashMap;
import java.util.HashSet;
import java.util.ArrayList;
import java.util.Scanner;

import antlr4.*;
import org.antlr.v4.runtime.*;
import org.antlr.v4.runtime.tree.ParseTree;

import intermediate.symtab.*;
import intermediate.type.*;
import backend.interpreter.*;

import static intermediate.symtab.SymtabEntry.Kind.*;
import static intermediate.type.Typespec.Form.*;

public class Commander extends CommanderBaseVisitor<Object>
{
    private RuntimeStack runtimeStack;
    private HashSet<Integer>     breakpoints = new HashSet<>();
    private HashSet<SymtabEntry> watchpoints = new HashSet<>();
    private Scanner console = new Scanner(System.in);
    boolean singleStepping  = false;
    
    /**
     * Constructor
     * @param runtimeStack the runtime stack.
     */
    public Commander(RuntimeStack runtimeStack)
    {
        this.runtimeStack = runtimeStack;
    }
    
    /**
     * Start reading commands.
     * @param ctx the current context.
     */
    public void start(ParserRuleContext ctx)
    {
        readCommands(ctx);
    }
    
    /**
     * Read a debugger command from the console.
     * @param ctx the current context.
     */
    private void readCommands(ParserRuleContext ctx)
    {
        System.out.println();
        printLineNumber(ctx.getStart().getLine());
        System.out.println();
        
        boolean resume = false;
        
        do
        {
            try
            {
                System.out.print("Command? ");
                String command = console.nextLine() + '\n';
                
                // Create the character stream from the input string.
                CharStream cs = CharStreams.fromString(command);
                
                // Create a lexer which scans the character stream
                // to create a token stream.
                CommanderLexer lexer = new CommanderLexer(cs);
                lexer.removeErrorListeners();
                CommonTokenStream tokens = new CommonTokenStream(lexer);
                
                // Create a parser which parses the token stream.
                CommanderParser parser = new CommanderParser(tokens);
                
                // Parse the command and visit the parse tree.
                ParseTree tree = parser.command();
                resume = (boolean) visit(tree);
            }
            catch (Exception ex)
            {
                System.out.println("*** Debugger command error.");
                resume = false;
            }
        } while (!resume);
    }

    /**
     * Debug a statement if single stepping or the statement is a breakpoint.
     * @param ctx the statement context.
     */
    public void processStatement(PascalParser.StatementContext ctx)
    {
        int lineNumber = ctx.getStart().getLine();
        
        if (singleStepping || breakpoints.contains(lineNumber))
        {
            readCommands(ctx);
        }
    }
    
    /**
     * Debug an assignment if the target variable is a watchpoint.
     * @param ctx the target variable context.
     * @param value the value to assign.
     */
    public void processAssignment(PascalParser.VariableContext ctx, 
                                  Object value, Typespec type)
    {
        if (watchpoints.contains(ctx.entry))
        {
            printLineNumber(ctx.getStart().getLine());
            System.out.println(": " + ctx.entry.getName() + " <= " + value);
        }
    }

    /**
     * Debug a variable factor if the variable is a watchpoint.
     * @param ctx the variable context.
     * @param value the variable's value.
     */
    public void processVariableFactor(PascalParser.VariableContext ctx, 
                                      Object value, Typespec type)
    {   
        if (watchpoints.contains(ctx.entry))
        {
            printLineNumber(ctx.getStart().getLine());
            System.out.println(": " + ctx.entry.getName() + " :: " + value);
        }
    }
    
    /**
     * Print the current line number.
     * @param lineNumber the line number.
     */
    private void printLineNumber(int lineNumber)
    {
        System.out.printf("At line %03d", lineNumber);
    }
    
    @Override
    public Object visitGoCommand(CommanderParser.GoCommandContext ctx)
    {
        singleStepping = false;        
        return true;
    }

    @Override
    public Object visitStepCommand(CommanderParser.StepCommandContext ctx)
    {
        singleStepping = true;        
        return true;
    }

    @Override
    public Object visitQuitCommand(CommanderParser.QuitCommandContext ctx)
    {
        System.out.println("Program terminated.");
        System.exit(0);
        
        return false;
    }

    @Override
    public Object visitBreakCommand(CommanderParser.BreakCommandContext ctx)
    {
        // Record the list of breakpoints.
        if (ctx.lineNumberList() != null)
        {
            for (CommanderParser.LineNumberContext lineNumberCtx : 
                                            ctx.lineNumberList().lineNumber())
            {
                int lineNumber = 
                            Integer.parseInt(lineNumberCtx.INTEGER().getText());
                breakpoints.add(lineNumber);
            }
        }
        else
        {
            System.out.print("Breakpoints:");
            
            // Print the list of breakpoints.
            if (breakpoints.size() > 0)
            {
                for (int lineNumber : breakpoints) 
                {
                    System.out.printf(" %03d", lineNumber);
                }
                System.out.println();
            }
            else
            {
                System.out.println(" none");
            }
        }
        
        return false;
    }

    @Override
    public Object visitUnbreakCommand(CommanderParser.UnbreakCommandContext ctx)
    {
        for (CommanderParser.LineNumberContext lineNumberCtx : 
                                            ctx.lineNumberList().lineNumber())
        {
            int lineNumber = Integer.parseInt(lineNumberCtx.INTEGER().getText());
            breakpoints.remove(lineNumber);
        }
        
        return false;
    }

    @Override
    public Object visitShowCommand(CommanderParser.ShowCommandContext ctx)
    {
        for (CommanderParser.VariableContext variableCtx :
                                                ctx.variableList().variable())
        {
            String variableName = variableCtx.IDENTIFIER().getText();
            
            // Print the variable name and the value.
            Cell variableCell = getCell(variableCtx);
            if (variableCell != null)
            {
                Object value = variableCell.getValue();
                System.out.println(variableName + " :: " + value);
            }
            else
            {
                System.out.println("*** Undeclared or invalid: " + variableName);
            }
        }
        
        return false;
    }
    
    /**
     * Get a variable's cell from its stack frame.
     * @param variableCtx the variable context.
     * @return the cell.
     */
    @SuppressWarnings("unchecked")
    private Cell getCell(CommanderParser.VariableContext variableCtx)
    {
        String variableName = variableCtx.IDENTIFIER().getText();
        int currentLevel = runtimeStack.currentNestingLevel();
        StackFrame frame = null;
        Cell cell = null;
        
        // Find the variable's cell in the runtime stack.
        for (int level = currentLevel; (cell == null) && (level > 0); level--)
        {
            frame = runtimeStack.getTopmost(level);
            cell = frame.getCell(variableName);
        }
        
        if (cell == null) return null;
        
        // Get the variable's datatype.
        SymtabEntry routineId = frame.getRoutineId();
        Symtab routineSymtab = routineId.getRoutineSymtab();
        SymtabEntry variableId = routineSymtab.lookup(variableName);
        Typespec variableType = variableId.getType();

        // Execute any array subscripts or record fields.
        for (CommanderParser.ModifierContext modCtx : variableCtx.modifier()) 
        {
            // Subscripts.
            if (modCtx.indexList() != null)
            {
                if (variableType.getForm() == ARRAY)
                {
                    // Compute a new reference for each subscript.
                    for (CommanderParser.IndexContext indexCtx :
                                                    modCtx.indexList().index())
                    {
                        Typespec indexType = variableType.getArrayIndexType();
                        int minIndex = 0;
                        
                        if (indexType.getForm() == SUBRANGE)
                        {
                            minIndex = indexType.getSubrangeMinValue();
                        }
                        
                        int value = 
                                Integer.parseInt(indexCtx.INTEGER().getText());
                        int index = value - minIndex;
                        
                        cell = ((Cell[]) cell.getValue())[index];
                        variableType = variableType.getArrayElementType();
                    }
                }
                else
                {
                    System.out.println("*** Not an array.");
                    return null;
                }
            }

            // Record field.
            else 
            {
                if (variableType.getForm() == RECORD)
                {
                    Symtab symtab = variableType.getRecordSymtab();
                    String fieldName = 
                                modCtx.field().IDENTIFIER().getText().toLowerCase();
                    SymtabEntry fieldId = symtab.lookup(fieldName);
                    
                    if (fieldId != null)
                    {
                        // Compute a new reference for the field.
                        HashMap<String, Cell> map =
                                        (HashMap<String, Cell>) cell.getValue();
                        cell = map.get(fieldName);
                        variableType = fieldId.getType();
                    }
                    else
                    {
                        System.out.println("*** Invalid field: " + fieldName);
                        return null;
                    }
                }
                else
                {
                    System.out.println("*** Not a record.");
                    return null;
                }
            }
        }

        return cell;
    }

    @Override
    public Object visitStackCommand(CommanderParser.StackCommandContext ctx)
    {
        ArrayList<SymtabEntry> callStack = new ArrayList<>();

        // Loop over the stack frames on the runtime stack
        // starting at the top of stack.
        for (int i = runtimeStack.size() - 1; i >= 0; --i) 
        {
            StackFrame frame = runtimeStack.get(i);
            SymtabEntry routineId = frame.getRoutineId();
            System.out.print((routineId.getKind() == FUNCTION ? "FUNCTION "
                                                              : "PROCEDURE "));
            System.out.println(routineId.getName() + ":");

            // Add the symbol table entry of the procedure or function.
            callStack.add(routineId);

            // Print the value of each local variable.
            for (String name : frame.getAllNames()) 
            {
                Object value = frame.getCell(name).getValue();
                System.out.println("  " + name + " :: " + value);
            }
        }
        
        return false;
    }

    @Override
    public Object visitWatchCommand(CommanderParser.WatchCommandContext ctx)
    {
        // Record the list of watchpoints.
        if (ctx.simpleVariableList() != null)
        {
            for (CommanderParser.SimpleVariableContext variableCtx :
                                    ctx.simpleVariableList().simpleVariable())
            {
                String variableName = variableCtx.IDENTIFIER().getText();
                SymtabEntry id = getSymtabEntry(variableName);
                
                if (id != null) watchpoints.add(id);
                else System.out.println("*** Undeclared: " + variableName);
            }
        }
        else
        {
            System.out.print("Watchpoints:");
            
            // Print the list of watchpoints.
            if (watchpoints.size() > 0)
            {
                for (SymtabEntry id : watchpoints)
                {
                    System.out.printf(" %s", id.getName());
                }
                System.out.println();
            }
            else
            {
                System.out.println(" none");
            }
        }
        
        return false;
    }

    @Override
    public Object visitUnwatchCommand(CommanderParser.UnwatchCommandContext ctx)
    {
        for (CommanderParser.SimpleVariableContext variableCtx :
                                    ctx.simpleVariableList().simpleVariable())
        {
            String variableName = variableCtx.IDENTIFIER().getText();
            SymtabEntry id = getSymtabEntry(variableName);
            
            if (id != null) watchpoints.remove(id);
            else System.out.println("*** Undeclared: " + variableName);
        }
        
        return false;
    }
    
    /**
     * Get the symbol table entry of a variable.
     * @param variableName the variable's name.
     * @return the symbol table entry.
     */
    private SymtabEntry getSymtabEntry(String variableName)
    {
        SymtabEntry variableId = null;
        int currentLevel = runtimeStack.currentNestingLevel();
        
        // Look for the stack frame.
        for (int level = currentLevel; 
             (variableId == null) && (level > 0); 
             level--)
        {
            // Get the symbol table of the stack frame's routine. 
            StackFrame frame = runtimeStack.getTopmost(level);
            SymtabEntry routineId = frame.getRoutineId();
            Symtab routineSymtab = routineId.getRoutineSymtab();
            
            // Look up the variable in the routine's symbol table.
            variableId = routineSymtab.lookup(variableName);
        }
        
        return variableId;
    }
}
